/* 
   cadaver, command-line DAV client
   Copyright (C) 1999, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
   $Id: cadaver.c,v 1.15 1999/11/24 17:05:13 joe Exp $
*/

#include <config.h>

#include <sys/types.h>

#include <sys/stat.h>

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif 
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <ctype.h>

/* joe: My man pages say getpass() should be in pwd.h. */
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

#include <signal.h>

#ifdef HAVE_READLINE_H
#include <readline.h>
#elif HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#endif

#ifdef HAVE_HISTORY_H
#include <history.h>
#elif HAVE_READLINE_HISTORY_H
#include <readline/history.h>
#endif

#include <errno.h>

#include "getopt.h"
#ifdef ENABLE_NETRC
#include "netrc.h"
#endif /* ENABLE_NETRC */

#include "frontend.h"
#include "protocol.h"
#include "httpdav.h"
#include "string_utils.h"
#include "basename.h"
#include "dirname.h"
#include "common.h"
#include "dates.h"

/* Local variables */
static bool child_running; /* true when we have a child running */

#ifdef ENABLE_NETRC
static netrc_entry *netrc_list; /* list of netrc entries */
#endif

static struct proto_host server; /* server details */
static char path[BUFSIZ] = {0}; /* current working collection */

static char *progname; /* argv[0] */

static bool have_connection; /* true when we have server details */

static const struct command *get_command( const char *name );

static void usage(void);
static void parse_args( int argc, char *argv[],
		       struct proto_host *server, char *path );
static char *read_command(void);
static char *choose_pager(void);

static void quit_handler( int signal );

static void count_lemons( void );

static void close_connection( void );
static void open_connection( char *hostname, const char *path );

static void init_signals( void );
static void init_netrc( void );

static char *resolve_path( const char *path, const char *filename, 
			   bool isdir );
static char *clever_path( const char *path, const char *src, 
			  const char *dest );

static void execute_get( const char *remote, const char *local );
static void execute_move( const char *from, const char *to );
static void execute_copy( const char *from, const char *to );
static void execute_mkcol( const char *filename );
static void execute_delete( const char *filename );
static void execute_cat( const char *filename );
static void execute_less( const char *filename );
static void execute_ls( const char *path );

static char transfer_msg[BUFSIZ];
enum transfer_status {
    tran_none, tran_starting, tran_going
} transfer_status;

enum command_id {
    cmd_ls, cmd_cd, cmd_quit, cmd_debug, cmd_open, cmd_close, cmd_lemon,
    cmd_help, cmd_put, cmd_get, cmd_mkcol, cmd_delete, cmd_move, cmd_copy,
    cmd_less, cmd_cat, cmd_unknown
};

/* Could do this with a separate command_name -> id map, would
 * be more elegant the the copy'n'paste for command aliases */
const static struct command {
    enum command_id id;
    const char *name;
    bool needs_connection;
    int min_arguments;
} commands[] = {
    { cmd_ls, "ls", true, 0 },
    { cmd_put, "put", true, 1 },
    { cmd_get, "get", true, 0 },
    { cmd_cat, "cat", true, 1 },
    { cmd_less, "less", true, 1 }, { cmd_less, "more", true, 1 },
    { cmd_mkcol, "mkcol", true, 1 }, { cmd_mkcol, "mkdir", true, 1 },
    { cmd_delete, "rm", true, 1 }, { cmd_delete, "delete", true, 1 },
    { cmd_copy, "copy", true, 2 }, { cmd_copy, "cp", true, 2 },
    { cmd_move, "move", true, 2 }, { cmd_move, "mv", true, 2 },
    { cmd_cd, "cd", true, 1 },
    { cmd_quit, "quit", false, 0 },
    { cmd_debug, "debug", false, 0 },
    { cmd_help, "help", false, 0 },
    { cmd_open, "open", false, 1 },
    { cmd_close, "close", true, 0 },
    { cmd_lemon, "", false, 0 },
    { cmd_unknown, NULL, 0 }
};    

static const char *help_string = 
"Commands: \n"
"  ls [path]                 List current [or other] collection contents\n"
"  cd path                   Change to specified collection\n"
"  open host[:port] [/path/] Open connection to server at path\n"
"  close                     Close current connection\n"
"  get remote [local]        Download remote resource\n"
"  put local [remote]        Upload local file\n"
"  mkcol remote              Create remote collection\n"
"  delete remote             Delete remote resource\n"
"  copy source dest          Copy resource from source to dest\n"
"  move source dest          Move resource from source to dest\n"
"  cat remote                Display remote resource\n"
"  less remote               Display remote resource through pager\n"
"  debug [x]                 Set (or just display) debugging level\n"
"  help                      Display this help message.\n"
"  quit                      Exit program.\n"
"Aliases:\n"
"  rm -> delete ; mkdir -> mkcol ; mv -> move ; cp -> copy ; more -> less\n"
"Resource name equivalences:\n"
"  . -> current collection\n";

static void usage() {
    printf( 
"Usage: %s [OPTIONS] hostname[:port] [path]\n"
"Options:\n"
"  -e, --expect100  Enable sending of `Expect: 100-continue' header.\n"
"     *** Warning: For Apache servers, use only with version 1.3.9 and later.\n"
"  -V, --version    Display version information.\n"
"  -h, --help       Display this help message.\n"
"Please send bug reports to Joe Orton <joe@orton.demon.co.uk>\n", progname );
}

static void set_path( const char *rel ) {
    char *newpath;
    newpath = resolve_path( path, rel, true );
    strcpy( path, newpath );
    free( newpath );
}
	
static void version( void ) {
    printf( PACKAGE " " VERSION "\n" );
}

static void close_connection( void ) {
    http_finish( );
    have_connection = false;
    strcpy( path, "!" );
    printf( "Connection closed.\n" );
}

static void open_connection( char *hostname, const char *path ) {
    char *pnt;
    const char *use_path;
    pnt = strchr( hostname, ':' );
    if( pnt != NULL && (strlen(pnt+1)>0) ) {
	*pnt = '\0';
	server.port = atoi( pnt+1 );
    } else {
	server.port = 80;
    }
    server.hostname = hostname;
#ifdef ENABLE_NETRC
    {
	netrc_entry *found;
	found = search_netrc( netrc_list, hostname );
	if( found != NULL ) {
	    if( found->account && found->password ) {
		server.username = found->account;
		server.password = found->account;
	    }
	}
    }
#endif /* ENABLE_NETRC */
    have_connection = false;
    /* Default to '/' if no path given */
    use_path = path?path:"/";
    switch( http_init( use_path, &server, NULL ) ) {
    case PROTO_OK:
	if( !http_webdav_server ) {
	    printf( "Collection `%s' is not DAV-enabled.\n", use_path );
	    close_connection( );
	} else {
	    set_path( use_path );
	    have_connection = true;
	}
	break;
    case PROTO_LOOKUP:
	printf( "Could not resolve hostname.\n" );
	break;
    case PROTO_CONNECT:
	printf( "Could not connect to remote host.\n" );
	break;
    default:
	printf( "Could not open connection:\n%s\n", http_error );
	break;
    }
}
	    
static void parse_args( int argc, char *argv[], 
		       struct proto_host *server, char *path ) {
    static const struct option opts[] = {
	{ "version", no_argument, NULL, 'V' },
	{ "help", no_argument, NULL, 'h' },
	{ "expect100", no_argument, NULL, 'e' },
	{ 0, 0, 0, 0 }
    };
    int optc;
    while( (optc = getopt_long( argc, argv, "ehV", opts, NULL )) != -1 ) {
	switch( optc ) {
	case 'h': usage( ); exit( -1 );
	case 'V': version( ); exit( -1 );
	case 'e': http_enable_expect = true; break;
	case '?': 
	default:
	    printf( "Try `%s --help' for more information.\n", progname );
	    exit(-1);
	}
    }
    if( argc == 1 ) {
	strcpy( path, "!" );
	have_connection = false;
    } else if( argc == 2 ) {
	open_connection( argv[1], NULL );
    } else if( argc == 3 ) {
	open_connection( argv[1], argv[2] );
    } else {
	usage( );
	exit( -1 );
    }
}

static char * read_command(void) {
    static char prompt[BUFSIZ];
    snprintf( prompt, BUFSIZ, "dav:%s> ", path );
    return readline( prompt ); 
}

static void execute_ls( const char *remote ) {
    int ret;
    char *real_remote;
    struct proto_file *files, *current;
    if( remote != NULL ) {
	real_remote = resolve_path( path, remote, true );
    } else {
	real_remote = strdup( path );
    }
    ret = dav_fetch( real_remote, 0, &files );
    if( ret == PROTO_OK ) {
	/* Easy this, eh? */
	if( files == NULL ) {
	    printf( "Collection `%s' is empty.\n", real_remote );
	} else {
	    printf( "Listing of collection `%s':\n", real_remote );
	    for( current = files; current!=NULL; current = current->next ) {
		char *date = rfc1123_date( current->modtime ), *restype;
		switch( current->type ) {
		case proto_file: restype = "Resource"; break;
		case proto_link: restype = "Reference"; break;
		case proto_dir: restype = "Collection"; break;
		default: restype = "Unknown"; break;
		}		    
		printf( "%10s: %-20s %10d  %s\n", restype, current->filename, 
			current->size, date );
		free( date );
	    }
	}
    } else {
	printf( "Could not retrieve listing for `%s':\n%s\n", 
		real_remote, http_error );
    }
    free( real_remote );
}

static void execute_mkcol( const char *filename ) {
    char *remote;
    remote = resolve_path( path, filename, true );
    (void) dav_mkcol( remote );
    printf( "%s\n", http_error );
    free( remote );
}

static void execute_delete( const char *filename ) {
    char *remote;
    remote = resolve_path( path, filename, false );
    (void) http_delete( remote );
    printf( "%s\n", http_error );
    free( remote );
}

/* Like resolve_path except more intelligent. */
static char *clever_path( const char *path, const char *src, 
			  const char *dest ) {
    char *ret;
    bool src_is_coll, dest_is_coll;
    dest_is_coll = (dest[strlen(dest)-1] == '/');
    src_is_coll = (src[strlen(src)-1] == '/');
    if( strcmp( dest, "." ) == 0 ) {
	ret = resolve_path( path, base_name( src ), false );
    } else if( strcmp( dest, ".." ) == 0 ) {
	char *path2, *parent;
	/* We need the parent of path... 
	 * Sadly, dir_name("/a/b/") == dir_name("/a/b/")
	 * So, dup path and strip off the trailing slash.
	 */
	path2 = strdup( path );
	path2[strlen(path2)-1] = '\0';
	parent = dir_name( path2 );
	ret = resolve_path( parent, base_name( src ), false );
	free( parent );
	free( path2 );
    } else if( !src_is_coll && dest_is_coll ) {
	/* Moving a file to a collection... the destination should
	 * be the basename of file concated with the collection. */
	char *tmp;
	tmp = resolve_path( path, dest, true );
	CONCAT2( ret, tmp, base_name( src ) );
    } else {
	ret = resolve_path( path, dest, false );
    }
    return ret;
}

static void cat_callback( void *userdata, const char *buf, const size_t len ) {
    FILE *f = userdata;
    if( fwrite( buf, len, 1, f ) < 0 ) {
	perror( "fwrite" );
    }
}

static char *choose_pager(void) {
    struct stat st;
    char *tmp;
    tmp = getenv("PAGER");
    if( tmp != NULL ) {
	return tmp;
    } else if( stat( "/usr/bin/less", &st ) == 0 ) {
	return "/usr/bin/less";
    } else if( stat( "/bin/less", &st ) == 0 ) {
	return "/bin/less";
    } else {
	return "/bin/more";
    }
}

static FILE *spawn_pager( const char *pager ) {
    /* Create a pipe */
    return popen( pager, "w" );
}

static void kill_pager( FILE *p ) {
    /* This blocks until the pager quits. */
    pclose( p );
}

static void execute_less( const char *resource ) {
    char *real_res;
    char *pager;
    FILE *p;
    real_res = resolve_path( path, resource, false );
    pager = choose_pager();
    printf( "Displaying `%s':\n", real_res );
    p = spawn_pager(pager);
    if( p == NULL ) {
	printf( "Error! Could not spawn pager `%s':\n%s\n", pager,
		strerror(errno) );
    } else {
	child_running = true;
	http_read_file( real_res, 0, cat_callback, p );
	kill_pager( p ); /* Blocks until the pager quits */
	child_running = false;
    }
}

static void execute_cat( const char *resource ) {
    char *real_res = resolve_path( path, resource, false );
    printf( "Displaying `%s':\n", real_res );
    (void) http_read_file( real_res, 0, cat_callback, stdout );
    printf( "%s\n", http_error );	
}

static void execute_move( const char *from, const char *to ) {
    char *real_from, *real_to;
    real_from = resolve_path( path, from, false );
    real_to = clever_path( path, from, to );
    printf( "Moving `%s' to `%s': \n", real_from, real_to );
    fflush( stdout );
    (void) dav_move( real_from, real_to );
    printf( "%s\n", http_error );
    free( real_from );
    free( real_to );
}

static void execute_copy( const char *from, const char *to ) {
    char *real_from, *real_to;
    real_from = resolve_path( path, from, false );
    real_to = clever_path( path, from, to );
    printf( "Copying `%s' to `%s': \n", real_from, real_to );
    fflush( stdout );
    (void) dav_copy( real_from, real_to );
    printf( "%s\n", http_error );
    free( real_from );
    free( real_to );
}

/* Returns absolute filename which is 'filename' relative to 
 * (absolute filename) 'path'. e.g.
 *    resolve_path( "/dav/", "asda" ) == "/dav/asda"
 *    resolve_path( "/dav/", "/asda" ) == "/asda"
 * Also removes '..' segments, e.g.
 *    resolve_path( "/dav/foobar/", "../fish" ) == "/dav/fish"
 * If isdir is true, ensures the return value has a trailing slash.
 */
static char *resolve_path( const char *path, const char *filename,
			   bool isdir ) {
    char *ret, *pnt;
    if( *filename == '/' ) {
	/* It's absolute */
	ret = strdup( filename );
    } else if( strcmp( filename, "." ) == 0 ) {
	ret = strdup( path );
    } else {
	CONCAT2( ret, path, filename );
    }
    if( isdir && ret[strlen(ret)-1] != '/' ) {
	char *newret;
	CONCAT2( newret, ret, "/" );
	free( ret );
	ret = newret;
    }
    /* Sort out '..', etc... */
    do {
	pnt = strstr( ret, "/../" );
	if( pnt != NULL ) {
	    char *last;
	    /* Find the *previous* path segment, to overwrite...
	     *    /foo/../
	     *    ^- last points here */
	    if( pnt > ret ) {
		for( last = pnt-1; (last > ret) && (*last != '/'); last-- );
	    } else {
		last = ret;
	    }
	    memmove( last, pnt + 3, strlen(pnt+2) );
	} else {
	    pnt = strstr( ret, "/./" );
	    if( pnt != NULL ) {
		memmove( pnt, pnt+2, strlen(pnt+1) );
	    }
	}
    } while( pnt != NULL );
    return ret;    
}

void execute_get( const char *remote, const char *local ) {
    char *filename, *real_remote;
    if( local == NULL ) {
	struct stat st;
	/* Choose an appropriate local filename */
	if( stat( base_name(remote), &st ) == 0 ) {
	    /* File already exists... don't overwrite */
	    filename = readline( "Enter local filename: " );
	} else {
	    filename = strdup( base_name( remote ) );
	}
    } else {
	filename = strdup( local );
    }
    real_remote = resolve_path( path, remote, false );
    /* Muck around like this because otherwise the connection status
     * messages and username/password prompts would mess up the nice
     * progress dots. */
    snprintf( transfer_msg, BUFSIZ, 
	      "Downloading `%s' to %s [", real_remote, filename );
    transfer_status = tran_starting;
    switch( http_get( filename, real_remote, 0, false ) ) {
    case PROTO_OK:
	if( transfer_status == tran_starting ) {
	    printf( transfer_msg );
	}
	printf( "] done.\n" );
	break;
    default:

	printf( "Failed:\n%s\n", http_error );
	break;
    }
    transfer_status = tran_none;
    free( real_remote );
    free( filename );
}

static void execute_put( const char *local, const char *remote ) {
    char *real_remote;
    if( remote == NULL ) {
	real_remote = resolve_path( path, base_name( local ), false );
    } else {
	real_remote = resolve_path( path, remote, false );
    }
    snprintf( transfer_msg, BUFSIZ,
	      "Uploading %s to `%s' [", local, real_remote );
    transfer_status = tran_starting;
    switch( http_put( local, real_remote, false ) ) {
    case PROTO_OK:
	if( transfer_status == tran_starting ) {
	    printf( transfer_msg );
	}
	printf( "] done.\n" );
	break;
    default:
	printf( "Failed:\n%s\n", http_error );
	break;
    }
    transfer_status = tran_none;
    free( real_remote );
}

const struct command *get_command( const char *name ) {
    int n;
    for( n = 0; commands[n].name != NULL; n++ ) {
	if( strcasecmp( commands[n].name, name ) == 0 ) {
	    return &commands[n];
	}
    }
    return NULL;
}

static void count_lemons( void ) {
    static int lemon_counter = 0;
    if( (++lemon_counter % 15) == 0 ) {
	printf( "Type something, I'm getting bored of this!\n" );
    }
}

static int execute_command( const char *line ) {
    char **tokens;
    const struct command *cmd;
    tokens = split_string( line, ' ', NULL, NULL );
    if( tokens[0] == NULL ) return 0;
    cmd = get_command( tokens[0] );
    if( cmd == NULL ) {
	printf( "You what? Try `help'.\n" );
	return 0;
    }
    if( (tokens[1] == NULL) && (cmd->min_arguments == 1) ) {
	printf( "Argument required, see `help' for details.\n" );
	return 0;
    } else if( (tokens[2] == NULL) && (cmd->min_arguments == 2) ) {
	printf( "Two arguments required, see `help' for details.\n" );
	return 0;
    }
    if( !have_connection && cmd->needs_connection ) {
	printf( "No connection - try `open' first.\n" );
	return 0;
    }
    switch( cmd->id ) {
    case cmd_debug:
	if( tokens[1] != NULL )
	    debug_mask = atoi( tokens[1] );
	printf( "Debugging level now: %d\n", debug_mask );
	break;
    case cmd_quit:
	return -1;
    case cmd_help:
	printf( help_string );
	break;
    case cmd_lemon:
	count_lemons();
	break;
    case cmd_less:
	execute_less( tokens[1] );
	break;
    case cmd_cat:
	execute_cat( tokens[1] );
	break;
    case cmd_cd:
	set_path( tokens[1] );
	break;
    case cmd_move:
	execute_move( tokens[1], tokens[2] );
	break;
    case cmd_copy:
	execute_copy( tokens[1], tokens[2] );
	break;
    case cmd_put:
	execute_put( tokens[1], tokens[2] );
	break;
    case cmd_get:
	execute_get( tokens[1], tokens[2] );
	break;
    case cmd_open:
	if( have_connection ) {
	    close_connection( );
	}
	open_connection( tokens[1], tokens[2] );
	break;
    case cmd_mkcol:
	execute_mkcol( tokens[1] );
	break;
    case cmd_delete:
	execute_delete( tokens[1] );
	break;
    case cmd_close:
	close_connection( );
	break;
    case cmd_ls:
	execute_ls( tokens[1] );
	break;
    case cmd_unknown:
	break;
    } 
    return 0;
}

static void quit_handler( int sig ) {
    /* Reinstall handler */
    if( child_running ) {
	/* The child gets the signal anyway... it can deal with it.
	 * Proper way is probably to ignore signals while child is
	 * running? */
	signal( sig, quit_handler );
	return;
    } else {
	printf( "Terminated by signal %d.\n", sig );
	if( have_connection ) {
	    close_connection( );
	}
	exit( -1 );
    }
}

static void init_signals( void ) {
    signal( SIGPIPE, SIG_IGN );
    signal( SIGTERM, quit_handler );
    signal( SIGABRT, quit_handler );
    signal( SIGQUIT, quit_handler );
    signal( SIGINT, quit_handler );
}

static void init_netrc( void ) {
#ifdef ENABLE_NETRC
    char *netrc;
    CONCAT2( netrc, getenv("HOME"), "/.netrc" );
    netrc_list = parse_netrc( netrc );
#endif
}

int main( int argc, char *argv[] ) {
    int ret = 0;
    progname = argv[0];
    debug_mask = 0;

    init_signals();
    init_netrc();

    parse_args( argc, argv, &server, path );

    while(ret == 0) {
	char *cmd;
	cmd = read_command();
	if( cmd == NULL ) {
	    ret = 1;
	} else {
#ifdef HAVE_ADD_HISTORY
	    if( strcmp( cmd, "" ) != 0 ) add_history( cmd );
#endif
	    ret = execute_command(cmd);
	    free( cmd );
	}
    }
    if( have_connection ) {
	close_connection( );
    }
    return 0;
}

void fe_connection( fe_conn_status status ) {
    if( transfer_status == tran_going ) return;
    switch( status ) {
    case fe_namelookup:
	printf( "Looking up hostname...\n" );
	break;
    case fe_connecting:
	printf( "Connecting to server...\n" );
	break;
    case fe_connected:
	printf( "Connected.\n" );
	break;
    }	    
}

void fe_transfer_progress( size_t progress, size_t total ) {
    if( transfer_status == tran_none ) return;
    if( transfer_status == tran_starting ) {
	printf( transfer_msg );
    }
    transfer_status = tran_going;
    putchar( '.' );
    fflush( stdout );
}

void fe_login( const char **username, const char **password ) {
    if( transfer_status == tran_going ) {
	printf( "] authentication failed.\n" );
    }
    *username = readline( "Username: " );
    *password = getpass( "Password: " );
    if( transfer_status == tran_going ) {
	printf( "Retrying [" );
	fflush( stdout );
    }
}
